import re
import warnings
from abc import ABC
from functools import cached_property
from itertools import product
from pathlib import Path

import numpy as np
import tifffile
import yaml
from ome_types import from_xml, model

from .. import AbstractReader, try_default


class Reader(AbstractReader, ABC):
    priority = 0
    do_not_pickle = "reader"

    @staticmethod
    def _can_open(path):
        if isinstance(path, Path) and path.suffix in (".tif", ".tiff"):
            with tifffile.TiffFile(path) as tif:
                return tif.is_imagej and tif.pages[-1]._nextifd() == 0  # noqa
        else:
            return False

    @cached_property
    def metadata(self):
        return {
            key: try_default(yaml.safe_load, value, value) if isinstance(value, str) else value
            for key, value in self.reader.imagej_metadata.items()
        }

    def get_ome(self):
        if self.reader.is_ome:
            match = re.match(r"^(.*)(pos.*)$", self.path.stem, flags=re.IGNORECASE)
            if match is not None and len(match.groups()) == 2:
                a, b = match.groups()
                with tifffile.TiffFile(self.path.with_stem(a + re.sub(r"\d", "0", b))) as file0:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=UserWarning)
                        ome = from_xml(file0.ome_metadata)
                        ome.images = [image for image in ome.images if self.path.stem[: len(image.name)] == image.name]
                        return ome

        page = self.reader.pages[0]
        size_y = page.imagelength
        size_x = page.imagewidth
        if self.p_ndim == 3:
            size_c = page.samplesperpixel
            size_t = self.metadata.get("frames", 1)  # // C
        else:
            size_c = self.metadata.get("channels", 1)
            size_t = self.metadata.get("frames", 1)
        size_z = self.metadata.get("slices", 1)
        if 282 in page.tags and 296 in page.tags and page.tags[296].value == 1:
            f = page.tags[282].value
            pxsize = f[1] / f[0]
        else:
            pxsize = None

        dtype = page.dtype.name
        if dtype not in (
            "int8",
            "int16",
            "int32",
            "uint8",
            "uint16",
            "uint32",
            "float",
            "double",
            "complex",
            "double-complex",
            "bit",
        ):
            dtype = "float"

        interval_t = self.metadata.get("interval", 0)

        ome = model.OME()
        ome.instruments.append(model.Instrument(id="Instrument:0"))
        ome.instruments[0].objectives.append(model.Objective(id="Objective:0"))
        ome.images.append(
            model.Image(
                id="Image:0",
                pixels=model.Pixels(
                    id="Pixels:0",
                    size_c=size_c,
                    size_z=size_z,
                    size_t=size_t,
                    size_x=size_x,
                    size_y=size_y,
                    dimension_order="XYCZT",
                    type=dtype,  # type: ignore
                    physical_size_x=pxsize,
                    physical_size_y=pxsize,
                ),
                objective_settings=model.ObjectiveSettings(id="Objective:0"),
            )
        )
        for c, z, t in product(range(size_c), range(size_z), range(size_t)):
            ome.images[0].pixels.planes.append(model.Plane(the_c=c, the_z=z, the_t=t, delta_t=interval_t * t))
        return ome

    def open(self):
        self.reader = tifffile.TiffFile(self.path)
        page = self.reader.pages.first
        self.p_ndim = page.ndim  # noqa
        if self.p_ndim == 3:
            self.p_transpose = [i for i in [page.axes.find(j) for j in "SYX"] if i >= 0]  # noqa
        else:
            self.p_transpose = [i for i in [page.axes.find(j) for j in "YX"] if i >= 0]  # noqa

    def close(self):
        self.reader.close()

    def __frame__(self, c: int, z: int, t: int):
        dimension_order = self.ome.images[0].pixels.dimension_order.value
        if self.p_ndim == 3:
            axes = "".join([ax.lower() for ax in dimension_order if ax.lower() in "zt"])
            ct = {"z": z, "t": t}
            n = sum([ct[ax] * np.prod(self.base_shape[axes[:i]]) for i, ax in enumerate(axes)])
            return np.transpose(self.reader.asarray(int(n)), self.p_transpose)[int(c)]
        else:
            axes = "".join([ax.lower() for ax in dimension_order if ax.lower() in "czt"])
            czt = {"c": c, "z": z, "t": t}
            n = sum([czt[ax] * np.prod(self.base_shape[axes[:i]]) for i, ax in enumerate(axes)])
        return np.transpose(self.reader.asarray(int(n)), self.p_transpose)



from abc import ABC
from itertools import product

import numpy as np
from ome_types import model

from .. import AbstractReader


class Reader(AbstractReader, ABC):
    priority = 20

    @staticmethod
    def _can_open(path):
        return isinstance(path, np.ndarray) and 1 <= path.ndim <= 5

    def get_ome(self):
        def shape(size_x=1, size_y=1, size_c=1, size_z=1, size_t=1):  # noqa
            return size_x, size_y, size_c, size_z, size_t

        size_x, size_y, size_c, size_z, size_t = shape(*self.array.shape)
        try:
            pixel_type = model.PixelType(self.array.dtype.name)
        except ValueError:
            if self.array.dtype.name.startswith("int"):
                pixel_type = model.PixelType("int32")
            else:
                pixel_type = model.PixelType("float")

        ome = model.OME()
        ome.instruments.append(model.Instrument())
        ome.images.append(
            model.Image(
                pixels=model.Pixels(
                    size_c=size_c,
                    size_z=size_z,
                    size_t=size_t,
                    size_x=size_x,
                    size_y=size_y,
                    dimension_order="XYCZT",
                    type=pixel_type,
                ),
                objective_settings=model.ObjectiveSettings(id="Objective:0"),
            )
        )
        for c, z, t in product(range(size_c), range(size_z), range(size_t)):
            ome.images[0].pixels.planes.append(model.Plane(the_c=c, the_z=z, the_t=t, delta_t=0))
        return ome

    def open(self):
        if isinstance(self.path, np.ndarray):
            self.array = np.array(self.path)
            while self.array.ndim < 5:
                self.array = np.expand_dims(self.array, -1)  # noqa
            self.path = "numpy array"

    def __frame__(self, c, z, t):
        frame = self.array[:, :, c, z, t]
        if self.axes.find("y") > self.axes.find("x"):
            return frame.T
        else:
            return frame



import re
from abc import ABC
from datetime import datetime
from itertools import product
from pathlib import Path

import tifffile
import yaml
from ome_types import model
from ome_types.units import _quantity_property  # noqa

from .. import AbstractReader


def lazy_property(function, field, *arg_fields):
    def lazy(self):
        if self.__dict__.get(field) is None:
            self.__dict__[field] = function(*[getattr(self, arg_field) for arg_field in arg_fields])
            try:
                self.model_fields_set.add(field)
            except Exception:  # noqa
                pass
        return self.__dict__[field]

    return property(lazy)


class Plane(model.Plane):
    """Lazily retrieve delta_t from metadata"""

    def __init__(self, t0, file, **kwargs):  # noqa
        super().__init__(**kwargs)
        # setting fields here because they would be removed by ome_types/pydantic after class definition
        setattr(self.__class__, "delta_t", lazy_property(self.get_delta_t, "delta_t", "t0", "file"))
        setattr(self.__class__, "delta_t_quantity", _quantity_property("delta_t"))
        self.__dict__["t0"] = t0  # noqa
        self.__dict__["file"] = file  # noqa

    @staticmethod
    def get_delta_t(t0, file):
        with tifffile.TiffFile(file) as tif:
            info = yaml.safe_load(tif.pages[0].tags[50839].value["Info"])
        return float((datetime.strptime(info["Time"], "%Y-%m-%d %H:%M:%S %z") - t0).seconds)


class Reader(AbstractReader, ABC):
    priority = 10

    @staticmethod
    def _can_open(path):
        pat = re.compile(r"(?:\d+-)?Pos.*", re.IGNORECASE)
        return (
            isinstance(path, Path)
            and path.is_dir()
            and (pat.match(path.name) or any(file.is_dir() and pat.match(file.stem) for file in path.iterdir()))
        )

    def get_ome(self):
        ome = model.OME()
        with tifffile.TiffFile(self.filedict[0, 0, 0]) as tif:
            metadata = {key: yaml.safe_load(value) for key, value in tif.pages[0].tags[50839].value.items()}
        ome.experimenters.append(
            model.Experimenter(id="Experimenter:0", user_name=metadata["Info"]["Summary"]["UserName"])
        )
        objective_str = metadata["Info"]["ZeissObjectiveTurret-Label"]
        ome.instruments.append(model.Instrument())
        ome.instruments[0].objectives.append(
            model.Objective(
                id="Objective:0",
                manufacturer="Zeiss",
                model=objective_str,
                nominal_magnification=float(re.findall(r"(\d+)x", objective_str)[0]),
                lens_na=float(re.findall(r"/(\d\.\d+)", objective_str)[0]),
                immersion=model.Objective_Immersion.OIL if "oil" in objective_str.lower() else None,
            )
        )
        tubelens_str = metadata["Info"]["ZeissOptovar-Label"]
        ome.instruments[0].objectives.append(
            model.Objective(
                id="Objective:Tubelens:0",
                manufacturer="Zeiss",
                model=tubelens_str,
                nominal_magnification=float(re.findall(r"\d?\d*[,.]?\d+(?=x$)", tubelens_str)[0].replace(",", ".")),
            )
        )
        ome.instruments[0].detectors.append(model.Detector(id="Detector:0", amplification_gain=100))
        ome.instruments[0].filter_sets.append(
            model.FilterSet(id="FilterSet:0", model=metadata["Info"]["ZeissReflectorTurret-Label"])
        )

        pxsize = metadata["Info"]["PixelSizeUm"]
        pxsize_cam = 6.5 if "Hamamatsu" in metadata["Info"]["Core-Camera"] else None
        if pxsize == 0:
            pxsize = pxsize_cam / ome.instruments[0].objectives[0].nominal_magnification
        pixel_type = metadata["Info"]["PixelType"].lower()
        if pixel_type.startswith("gray"):
            pixel_type = "uint" + pixel_type[4:]
        else:
            pixel_type = "uint16"  # assume

        size_c, size_z, size_t = (max(i) + 1 for i in zip(*self.filedict.keys()))
        t0 = datetime.strptime(metadata["Info"]["Time"], "%Y-%m-%d %H:%M:%S %z")
        ome.images.append(
            model.Image(
                pixels=model.Pixels(
                    size_c=size_c,
                    size_z=size_z,
                    size_t=size_t,
                    size_x=metadata["Info"]["Width"],
                    size_y=metadata["Info"]["Height"],
                    dimension_order="XYCZT",  # type: ignore
                    type=pixel_type,
                    physical_size_x=pxsize,
                    physical_size_y=pxsize,
                    physical_size_z=metadata["Info"]["Summary"]["z-step_um"],
                ),
                objective_settings=model.ObjectiveSettings(id="Objective:0"),
            )
        )

        for c, z, t in product(range(size_c), range(size_z), range(size_t)):
            ome.images[0].pixels.planes.append(
                Plane(
                    t0,
                    self.filedict[c, z, t],
                    the_c=c,
                    the_z=z,
                    the_t=t,
                    exposure_time=metadata["Info"]["Exposure-ms"] / 1000,
                )
            )

        # compare channel names from metadata with filenames
        pattern_c = re.compile(r"img_\d{3,}_(.*)_\d{3,}$", re.IGNORECASE)
        for c in range(size_c):
            ome.images[0].pixels.channels.append(
                model.Channel(
                    id=f"Channel:{c}",
                    name=pattern_c.findall(self.filedict[c, 0, 0].stem)[0],
                    detector_settings=model.DetectorSettings(
                        id="Detector:0", binning=metadata["Info"]["Hamamatsu_sCMOS-Binning"]
                    ),
                    filter_set_ref=model.FilterSetRef(id="FilterSet:0"),
                )
            )
        return ome

    def open(self):
        # /some_path/Pos4: path = /some_path, series = 4
        # /some_path/5-Pos_001_005: path = /some_path/5-Pos_001_005, series = 0
        if re.match(r"(?:\d+-)?Pos.*", self.path.name, re.IGNORECASE) is None:
            pat = re.compile(rf"^(?:\d+-)?Pos{self.series}$", re.IGNORECASE)
            files = sorted(file for file in self.path.iterdir() if pat.match(file.name))
            if len(files):
                path = files[0]
            else:
                raise FileNotFoundError(self.path / pat.pattern)
        else:
            path = self.path

        pat = re.compile(r"^img_\d{3,}.*\d{3,}.*\.tif$", re.IGNORECASE)
        filelist = sorted([file for file in path.iterdir() if pat.search(file.name)])
        with tifffile.TiffFile(self.path / filelist[0]) as tif:
            metadata = {key: yaml.safe_load(value) for key, value in tif.pages[0].tags[50839].value.items()}

        # compare channel names from metadata with filenames
        cnamelist = metadata["Info"]["Summary"]["ChNames"]
        cnamelist = [c for c in cnamelist if any([c in f.name for f in filelist])]

        pattern_c = re.compile(r"img_\d{3,}_(.*)_\d{3,}$", re.IGNORECASE)
        pattern_z = re.compile(r"(\d{3,})$")
        pattern_t = re.compile(r"img_(\d{3,})", re.IGNORECASE)
        self.filedict = {
            (
                cnamelist.index(pattern_c.findall(file.stem)[0]),  # noqa
                int(pattern_z.findall(file.stem)[0]),
                int(pattern_t.findall(file.stem)[0]),
            ): file
            for file in filelist
        }

    def __frame__(self, c=0, z=0, t=0):
        return tifffile.imread(self.path / self.filedict[(c, z, t)])



from .. import JVM

if JVM is None:
    __all__ = "cziread", "fijiread", "ndread", "seqread", "tifread", "metaseriesread"
else:
    __all__ = "bfread", "cziread", "fijiread", "ndread", "seqread", "tifread", "metaseriesread"



import re
from abc import ABC
from pathlib import Path
from typing import Optional

import tifffile
from ome_types import model
from ome_types.units import _quantity_property  # noqa

from .. import AbstractReader


class Reader(AbstractReader, ABC):
    priority = 20
    do_not_pickle = "last_tif"

    @staticmethod
    def _can_open(path):
        return isinstance(path, Path) and (
            path.is_dir() or (path.parent.is_dir() and path.name.lower().startswith("pos"))
        )

    @staticmethod
    def get_positions(path: str | Path) -> Optional[list[int]]:
        pat = re.compile(rf"s(\d)_t\d+\.(tif|TIF)$")
        return sorted({int(m.group(1)) for file in Path(path).iterdir() if (m := pat.search(file.name))})

    def get_ome(self):
        ome = model.OME()
        tif = self.get_tif(0)
        metadata = tif.metaseries_metadata
        size_z = len(tif.pages)
        page = tif.pages[0]
        shape = {axis.lower(): size for axis, size in zip(page.axes, page.shape)}
        size_x, size_y = shape["x"], shape["y"]

        ome.instruments.append(model.Instrument())

        size_c = 1
        size_t = max(self.filedict.keys()) + 1
        pixel_type = f"uint{metadata['PlaneInfo']['bits-per-pixel']}"
        ome.images.append(
            model.Image(
                pixels=model.Pixels(
                    size_c=size_c,
                    size_z=size_z,
                    size_t=size_t,
                    size_x=size_x,
                    size_y=size_y,
                    dimension_order="XYCZT",
                    type=pixel_type,
                ),
                objective_settings=model.ObjectiveSettings(id="Objective:0"),
            )
        )
        return ome

    def open(self):
        pat = re.compile(rf"s{self.series}_t\d+\.(tif|TIF)$")
        filelist = sorted([file for file in self.path.iterdir() if pat.search(file.name)])
        pattern = re.compile(r"t(\d+)$")
        self.filedict = {int(pattern.search(file.stem).group(1)) - 1: file for file in filelist}
        if len(self.filedict) == 0:
            raise FileNotFoundError
        self.last_tif = 0, tifffile.TiffFile(self.filedict[0])

    def close(self) -> None:
        self.last_tif[1].close()

    def get_tif(self, t: int = None):
        last_t, tif = self.last_tif
        if (t is None or t == last_t) and not tif.filehandle.closed:
            return tif
        else:
            tif.close()
            tif = tifffile.TiffFile(self.filedict[t])
            self.last_tif = t, tif
            return tif

    def __frame__(self, c=0, z=0, t=0):
        tif = self.get_tif(t)
        page = tif.pages[z]
        if page.axes.upper() == "YX":
            return page.asarray()
        elif page.axes.upper() == "XY":
            return page.asarray().T
        else:
            raise NotImplementedError(f"reading axes {page.axes} is not implemented")
